/*
 * This is part of the doxml library, a parser for XML.  It should not
 *  require any functionality not defined in ANSI C's standard libraries.
 *
 * Copyright (C) 1999 John Stracke, http://www.thibault.org
 *
 * This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public License
 *  as published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 * $Log: xml.h,v $
 * Revision 1.28  1999/07/24 21:58:30  francis
 * Implementing late binding for entity values, so that an entity can
 * reference another entity which is defined later.
 *
 * Revision 1.27  1999/07/24 19:13:31  francis
 * Adding support for notations
 *
 * Revision 1.26  1999/07/24 16:55:27  francis
 * Adding support for entity-valued attributes
 *
 * Revision 1.25  1999/07/24 15:23:03  francis
 * Adding support for default attribute values
 *
 * Revision 1.24  1999/07/24 13:13:39  francis
 * Fixing the problem where <D:foo> can be terminated by </foo> instead of
 * by </D:foo>.
 *
 * Revision 1.23  1999/07/19 11:35:18  francis
 * Documenting the changed data structures for namespace reform
 *
 * Revision 1.22  1999/07/19 01:31:11  francis
 * Implementing namespace reform (but not using it yet).  Not documented yet.
 *
 * Revision 1.21  1999/07/08 13:52:22  francis
 * Adding support for external parsed entities (I don't parse them yet,
 * but I do include an entity atom).  I'll hook in the URI framework
 * later.
 *
 * Fixing a bug in parse_markupdecl() where entities with no NDATA didn't
 * zero out notation_name.
 *
 * Revision 1.20  1999/06/25 21:17:37  francis
 * Changing DOXML_XML to DOXML_XML; adding idents
 *
 * Revision 1.19  1999/06/25 21:15:00  francis
 * Adding a framework for the application to provide doxml with URI
 * handlers for use in processing external IDs.  (Not providing URI
 * code as part of doxml means we can work with whatever URI code the
 * app wants, and don't have to write anything non-portable.)
 *
 * Also, changing old _LIBXML_ stuff in the headers to _DOXML_.
 *
 * Revision 1.18  1999/06/25 19:43:02  francis
 * Adding support for unparsed entities
 *
 * Revision 1.17  1999/06/21 03:19:56  francis
 * Finishing up <!ENTITY> support
 *
 * Revision 1.16  1999/06/21 02:53:28  francis
 * Adding the ability to parse <!ENTITY>s.  Can't write them yet, though.
 *
 * Revision 1.15  1999/06/21 00:59:30  francis
 * Adding support for <!ATTLIST>s
 *
 * Revision 1.14  1999/06/20 22:23:54  francis
 * Adding support for <!ELEMENT>s in the doctype
 *
 * Revision 1.13  1999/06/20 20:05:34  francis
 * Adding support for PIs in the doctype
 *
 * Revision 1.12  1999/06/20 16:30:14  francis
 * Adding partial support for <!DOCTYPE>.  It now handles doctypes with
 * names (all doctypes have names), with external-ids, but not with
 * any of the declarations.
 *
 * Revision 1.11  1999/06/13 05:12:45  francis
 * Correcting the support in the data structures for the doctype declaration
 *
 * Revision 1.10  1999/06/13 05:07:06  francis
 * Adding support in the data structures for the doctype declaration
 *
 * Revision 1.9  1999/06/07 04:57:38  francis
 * Implementing processing instructions.  Note the API changes.
 *
 * Revision 1.8  1999/06/01 02:44:43  francis
 * Changing doxml_tags_match() so that namespaces come before tags, as in
 * other functions.
 *
 * Revision 1.7  1999/06/01 01:51:14  francis
 * Adding functions to get/set/clear attributes.  Reorganzing the
 * to-do list.  Fixing a parser bug introduced I removed support for
 * attributes with no values at release 0.2.
 *
 * Revision 1.6  1999/05/31 16:05:27  francis
 * Clarifying the formatting comment on doxml_write().
 *
 * Revision 1.5  1999/05/31 14:43:39  francis
 * Removing the support for attributes without values (which turn out to
 * be illegal).
 *
 * Revision 1.4  1999/05/21 23:10:24  francis
 * Reformatting, since renaming to doxml changed the widths of the
 * identifiers.
 *
 * Revision 1.3  1999/05/21 23:05:41  francis
 * Finished renaming to doxml
 *
 * Revision 1.2  1999/05/21 22:59:48  francis
 * Partially renaming to doxml
 *
 * Revision 1.1.1.1  1999/05/21 22:20:36  francis
 * Initial checkin of libxml, to be renamed doxml
 *
 */
#ifndef _DOXML_DOXML_XML_H_
#define _DOXML_DOXML_XML_H_

static const char* _doxml_version_ident_doxml_doxml_h="$Id: xml.h,v 1.28 1999/07/24 21:58:30 francis Exp $";

#include <stdio.h>

/* Declare the opaque struct that represents a parsing session.  An
 *  doxml_context is constructed by a function that populates it with an
 *  doxml_getc_method (below) and data appropriate to that method.
 */
struct doxml_context_tag;
typedef struct doxml_context_tag doxml_context;

struct doxml_attribute_tag;
typedef struct doxml_attribute_tag doxml_attribute;

typedef struct {
  const char* name;
  const char* space;
  const char* prefix;
} doxml_name;

typedef struct {
  const char* name;
  const char* prefix;
} doxml_name_in_dtd;

struct doxml_markupdecl_tag;
typedef struct doxml_markupdecl_tag doxml_markupdecl;

struct doxml_attribute_tag {
  doxml_name name;

  int is_entity;

  union {
    const char* literal;
    const doxml_markupdecl* entity;
  } value;

  doxml_attribute* next;
  doxml_attribute* prev;
};

struct doxml_atom_tag;
typedef struct doxml_atom_tag doxml_atom;

typedef enum {
  doxml_atom_text,
  doxml_atom_element,
  doxml_atom_pi,
  doxml_atom_external_entity
} doxml_atom_type;

struct doxml_atom_tag {

  /* The atom represents either a sequence of text or an XML
   *  element or processing instruction.
   */

  doxml_atom_type type;

  union {
    struct {
      doxml_name name;
      struct {
        doxml_attribute* first;
        doxml_attribute* last;
      } attrs;

      struct {
        doxml_atom* first;
        doxml_atom* last;
      } atoms;
    } element;
    
    struct {
      const char* text;
    } text;
    
    struct {
      doxml_name name;
      const char* value;
    } pi;

    struct {
      /* Points to a doxml_markupdecl which represents the declaration of
       *  the entity.  decl->type is doxml_markupdecl_entity.
       *  decl->entity->name is the name of the entity.
       *  decl->entity->is_general is true.
       *  decl->entity->general->is_external is true.
       *  decl->entity->general->external.external_id is the external ID
       *  (SYSTEM and optional PUBLIC identifiers).  If the entity is
       *  unparsed, then decl->entity->general->external.notation_name
       *  is the name of the notation.  If the entity is parsed, then
       *  decl->entity->general->parsed should be the text obtained by
       *  fetching the definition (from the URI
       *  decl->entity->general->external.external_id.system_uri) and
       *  parsing it.  However, at this time (doxml 0.7.4), this fetch
       *  is not done.
       */
      const doxml_markupdecl* decl;
    } external_entity;
  } data;

  /* parent points to the atom's enclosing element.  If parent is
   *  NULL, then the atom is the document's root element.
   */
  doxml_atom* parent;
  doxml_atom* next;
  doxml_atom* prev;
};

typedef enum {
  doxml_markupdecl_element_empty,
  doxml_markupdecl_element_any,
  doxml_markupdecl_element_mixed,
  doxml_markupdecl_element_children
} doxml_markupdecl_element_type;

typedef enum {
  doxml_markupdecl_element,
  doxml_markupdecl_attlist,
  doxml_markupdecl_entity,
  doxml_markupdecl_notation,
  doxml_markupdecl_pi,
  doxml_markupdecl_pe_reference
} doxml_markupdecl_type;

struct doxml_name_node_tag;
typedef struct doxml_name_node_tag doxml_name_node;

struct doxml_name_node_tag {
  doxml_name_node* next;
  doxml_name_node* prev;
  doxml_name_in_dtd name;
};

typedef enum {
  doxml_multiplicity_1,         /* no notation */
  doxml_multiplicity_1_or_0,    /* ? */
  doxml_multiplicity_0_or_more, /* * */
  doxml_multiplicity_1_or_more  /* + */
} doxml_multiplicity;

struct doxml_markupdecl_element_cp_tag;
typedef struct doxml_markupdecl_element_cp_tag
doxml_markupdecl_element_cp;

typedef enum {
  doxml_markupdecl_element_cp_invalid,
  doxml_markupdecl_element_cp_choice,
  doxml_markupdecl_element_cp_seq,
  doxml_markupdecl_element_cp_name
} doxml_markupdecl_element_cp_type;

struct doxml_markupdecl_element_cp_tag {
  doxml_markupdecl_element_cp* next;
  doxml_markupdecl_element_cp* prev;

  doxml_markupdecl_element_cp_type type;
  doxml_multiplicity multiplicity;
  union {
    doxml_name_in_dtd name;
    doxml_markupdecl_element_cp* choice;
    doxml_markupdecl_element_cp* seq;
  } data;
};

struct doxml_markupdecl_attdef_tag;

typedef struct doxml_markupdecl_attdef_tag doxml_markupdecl_attdef;

typedef enum {
  doxml_markupdecl_attdef_attype_cdata,
  doxml_markupdecl_attdef_attype_id,
  doxml_markupdecl_attdef_attype_idref,
  doxml_markupdecl_attdef_attype_idrefs,
  doxml_markupdecl_attdef_attype_entity,
  doxml_markupdecl_attdef_attype_entities,
  doxml_markupdecl_attdef_attype_nmtoken,
  doxml_markupdecl_attdef_attype_nmtokens,
  doxml_markupdecl_attdef_attype_enumerated
} doxml_markupdecl_attdef_attype_type;

typedef enum {
  doxml_markupdecl_attdef_default_required,
  doxml_markupdecl_attdef_default_implied,
  doxml_markupdecl_attdef_default_fixed,
  doxml_markupdecl_attdef_default_plain
} doxml_markupdecl_attdef_default_type;

struct doxml_markupdecl_attdef_tag {
  doxml_markupdecl_attdef* next;
  doxml_markupdecl_attdef* prev;

  doxml_name_in_dtd name;

  struct {
    doxml_markupdecl_attdef_attype_type type;

    struct {
      doxml_name_node* names;
      int is_notation;
    } enumeration;
  } attype;

  struct {
    doxml_markupdecl_attdef_default_type type;
    const char* value;
  } def;
};

typedef struct {
  int is_public;
  const char* system_uri;
  const char* pubid;

  const char* uri_contents;
} doxml_external_id;

struct doxml_markupdecl_tag {
  doxml_markupdecl* next;
  doxml_markupdecl* prev;

  doxml_markupdecl_type type;

  union {
    struct {
      doxml_markupdecl_element_type type;

      doxml_name_in_dtd name;

      union {
        doxml_name_node* mixed;
        doxml_markupdecl_element_cp* children;
      } data;
    } element;

    struct {
      doxml_name_in_dtd name;
      doxml_markupdecl_attdef* attdefs;
    } attlist;

    struct {
      doxml_name_in_dtd name;
      int is_general;
      union {
        struct {
          int is_external;

          const char* parsed; /* What it parses down to */
          int parsing; /* Are we trying to parse it right now?
                        *  Used to detect circular references.
                        */

          union {
            const char* internal;

            struct {
              doxml_external_id external_id;
              doxml_name_in_dtd notation_name;
            } external;
          } data;

        } general;

        struct {
          int is_external;

          union {
            const char* internal;
            doxml_external_id external_id;
          } data;
        } param;

      } data;
    } entity;

    struct {
      doxml_name_in_dtd name;
      doxml_external_id external_id;
    } notation;

    struct {
      doxml_name_in_dtd name;
      const char* value;
    } pi;

    struct {
      doxml_name_in_dtd name;
    } pe_reference;
  } data;
};

typedef struct {
  doxml_name_in_dtd name;
  doxml_external_id* external_id;
  struct {
    doxml_markupdecl* first;
    doxml_markupdecl* last;
  } decls;
} doxml_doctype;

typedef struct {
  struct {
    const char* version;
    const char* encoding;
    int standalone; /* trivalue: 0 false, 1 true, -1 unspecified */
  } xml_decl;

  doxml_doctype* doctype;

  doxml_atom* prolog;

  doxml_atom* root;
} doxml_document;

/* Create an doxml_context that reads from stream.  If
 *  close_stream_after_parsing is nonzero, the stream will be closed
 *  when the doxml_context is closed.  (The reason one might not want
 *  to do this is that there might be XML embedded in a larger file.)
 *
 * Returns NULL on error, or a pointer to a freshly allocated doxml_context
 *  on success.
 *
 * The stream may contain 8-bit characters.
 */
doxml_context* doxml_open_stream(FILE* stream,
                                 int close_stream_after_parsing);

/* Open the file named path and create an doxml_context to parse
 *  it.  Naturally, this is a slim wrapper around doxml_open_stream(),
 *  above.
 *
 * The file will be opened for reading only.
 *
 * Returns NULL on error, or a pointer to a freshly allocated doxml_context
 *  on success.
 */
doxml_context* doxml_open_file(const char* path);

/* Create an doxml_context to parse the string str (which may contain 8-bit
 *  characters).  If free_str_after_parsing is nonzero, the string will be
 *  free()d when the doxml_context is closed.
 */
doxml_context* doxml_open_string(const char* str,
                                 int free_str_after_parsing);

/* Close the doxml_context: free() any private data, close any open streams
 *  (if appropriate), free() the context.
 */
void doxml_close(doxml_context* context);

/* Parse a document from context.  Returns NULL on error, or a freshly
 *  allocated doxml_document on success.
 *
 * The document will be constructed to be independent of context; that is,
 *  the caller can safely close context and still use the atom.
 *
 * If the parser encountered EOF without finding any non-whitespace text,
 *  the atom will be a text atom containing the empty string.
 */
doxml_document* doxml_parse(doxml_context* context);

/* free() the atom and all its components.
 */
void doxml_delete_atom(doxml_atom* atom);

/* free() the document and all its components.
 */
void doxml_delete_document(doxml_document* document);

/* Write atom to stream.  Returns 0 on success, or -1 on error.
 *
 * Note that the prettyprinting done is aimed more at data-oriented
 *  documents rather than text-oriented documents.
 */
int doxml_write_atom(FILE* stream,
                     doxml_atom* atom);

/* Write document to stream.  Returns 0 on success, or -1 on error.
 *
 * Note that the prettyprinting done is aimed more at data-oriented
 *  documents rather than text-oriented documents.
 */
int doxml_write_document(FILE* stream,
                         doxml_document* document);

/* Get the value of the attribute named name, in the namespace namespace,
 *  from the atom element (which, of course, must actually be an element,
 *  not a text sequence).  Returns NULL on error or the value found (which
 *  is not a fresh copy; the caller does not need to free() it).
 *
 * If the attribute is present, but has an entity value, returns NULL.  Yes,
 *  this is a problem.
 */
const char* doxml_attr_get(const doxml_atom* element,
                           const char* space,
                           const char* name);

/* Set the value of the attribute named name, in the namespace space,
 *  in the atom element (which, of course, must actually be an element,
 *  not a text sequence).  Returns 0 on success or <0 on error.
 */
int doxml_attr_set(doxml_atom* element,
                   const char* space,
                   const char* name,
                   const char* value);

/* Clear the value of the attribute named name, in the namespace space,
 *  in the atom element (which, of course, must actually be an element,
 *  not a text sequence).  Returns 0 on success or <0 on error.
 */
int doxml_attr_clear(doxml_atom* element,
                     const char* space,
                     const char* name);

/* Returns 1 (true) if name1 and name2 are equivalent names.  This is true
 *  if:
 *  * name{1,2} must not be NULL.
 *  * name{1,2}->name must not be NULL.
 *  * name{1,2}->name are equal strings.
 *  * name{1,2}->space are either both NULL or equal strings.
 *  * If prefixes_must_match is true, name{1,2}->prefix are either both
 *    NULL or equal strings.
 *
 * If second_is_terminator is true, then, instead of comparing name1->name
 *  to name2->name, it compares name1->name to (name2->name)+1, and
 *  checks that (name2->name)[0] is '/'.
 */
int doxml_tags_match(const doxml_name* name1,
                     const doxml_name* name2,
                     int prefixes_must_match,
                     int second_is_terminator);

int doxml_tags_match_dtd(const doxml_name* name_in_doc,
                         const doxml_name_in_dtd* name_in_dtd);

int doxml_tags_match_dtd2(const doxml_name_in_dtd* name1,
                          const doxml_name_in_dtd* name2);

#endif /*_DOXML_DOXML_XML_H_*/
