/*
 * This is part of the doxml library, a parser for XML.  It should not
 *  require any functionality not defined in ANSI C's standard libraries.
 *
 * Copyright (C) 1999 John Stracke, http://www.thibault.org
 *
 * This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public License
 *  as published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 * $Log: addattr.c,v $
 * Revision 1.6  1999/07/31 15:16:20  francis
 * Fixing a debug
 *
 * Revision 1.5  1999/07/24 16:55:28  francis
 * Adding support for entity-valued attributes
 *
 * Revision 1.4  1999/07/24 15:23:03  francis
 * Adding support for default attribute values
 *
 * Revision 1.3  1999/07/19 01:31:11  francis
 * Implementing namespace reform (but not using it yet).  Not documented yet.
 *
 * Revision 1.2  1999/06/07 04:57:38  francis
 * Implementing processing instructions.  Note the API changes.
 *
 * Revision 1.1  1999/06/01 01:51:14  francis
 * Adding functions to get/set/clear attributes.  Reorganzing the
 * to-do list.  Fixing a parser bug introduced I removed support for
 * attributes with no values at release 0.2.
 *
 */

static const char* id="$Id: addattr.c,v 1.6 1999/07/31 15:16:20 francis Exp $";

#include <doxml/priv.h>
#include <string.h>

static const char*
safestr(const char* s)
{
  return (s ? s : "(null)");
}

static doxml_attribute*
doxml_attr_add_with_prefix_no_value(doxml_atom* element,
                                    const char* space,
                                    const char* prefix,
                                    const char* name);

doxml_attribute*
doxml_attr_add(doxml_atom* element,
               const char* space,
               const char* name,
               const char* value)
{
  return doxml_attr_add_with_prefix(element,space,0,name,value);
}

doxml_attribute*
doxml_attr_add_with_prefix(doxml_atom* element,
                           const char* space,
                           const char* prefix,
                           const char* name,
                           const char* value)
{
  doxml_attribute* res=doxml_attr_add_with_prefix_no_value(element,
                                                           space,
                                                           prefix,
                                                           name);
  res->is_entity=0;
  res->value.literal=value;
  return res;
}

static doxml_attribute*
doxml_attr_add_with_prefix_no_value(doxml_atom* element,
                                    const char* space,
                                    const char* prefix,
                                    const char* name)

{
  doxml_attribute* res;

#ifdef DEBUG
  fprintf(stderr,"doxml_attr_add_with_prefix_no_value(%s,%s,%s)\n",
          safestr(space),
          safestr(prefix),
          safestr(name)
          );
#endif

  if (!element)
    {
      fprintf(stderr,"doxml_attr_add(): element is NULL\n");
      return 0;
    }

  if (!name)
    {
      fprintf(stderr,"doxml_attr_add(): name is NULL\n");
      return 0;
    }

  if (space)
    {
      space=strdup(space);
      if (!space)
        {
          fprintf(stderr,"doxml_attr_add(): could not strdup(space)\n");
          return 0;
        }
    }
  else
    {
      if (prefix)
        {
          space=doxml_lookup_namespace(element,prefix);
          if (space)
            {
              space=strdup(space);
              if (!space)
                {
                  fprintf(stderr,"doxml_attr_add(): could not strdup(space)\n");
                  return 0;
                }
            }
        }
    }

  res=(doxml_attribute*)malloc(sizeof(doxml_attribute));
  if (!res)
    {
      fprintf(stderr,"doxml_attr_add(): malloc() failed\n");
      if (space)
        free(space);
      return 0;
    }

  res->name.space=space;
  res->name.prefix=prefix;
  res->name.name=name;

  if (res->prev=element->data.element.attrs.last)
    res->prev->next=res;
  else
    element->data.element.attrs.first=res;

  res->next=0;
  element->data.element.attrs.last=res;

  return res;
}

void doxml_attr_del(doxml_atom* element,
                    doxml_attribute* attr)
{
  if (!element)
    {
      fprintf(stderr,"doxml_attr_del(): element==NULL\n");
      return;
    }

  if ((element->type)!=doxml_atom_element)
    {
      fprintf(stderr,"doxml_attr_del(): atom is not an element\n");
      return;
    }

  if (!attr)
    {
      fprintf(stderr,"doxml_attr_del(): attr==NULL\n");
      return;
    }

  if (attr->next)
    attr->next->prev=attr->prev;
  else
    element->data.element.attrs.last=attr->prev;

  if (attr->prev)
    attr->prev->next=attr->next;
  else
    element->data.element.attrs.first=attr->next;

  if (attr->name.space)
    free(attr->name.space);
  if (attr->name.name)
    free(attr->name.name);
  if (!(attr->is_entity))
    {
      if (attr->value.literal)
        free(attr->value.literal);
    }
  free(attr);
}

doxml_attribute* doxml_attr_add_entity(doxml_atom* atom,
                                       const char* space,
                                       const char* name,
                                       const doxml_markupdecl* value)
{
  return doxml_attr_add_entity_with_prefix(atom,space,0,name,value);
}

/* If space is 0, it will be deduced from prefix */
doxml_attribute* doxml_attr_add_entity_with_prefix(doxml_atom* atom,
                                                   const char* space,
                                                   const char* prefix,
                                                   const char* name,
                                                   const doxml_markupdecl* value)
{
  doxml_attribute* res=doxml_attr_add_with_prefix_no_value(atom,
                                                           space,
                                                           prefix,
                                                           name);
  res->is_entity=1;
  res->value.entity=value;
  return res;
}
