/*
 * This is part of the doxml library, a parser for XML.  It should not
 *  require any functionality not defined in ANSI C's standard libraries.
 *
 * Copyright (C) 1999 John Stracke, http://www.thibault.org
 *
 * This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public License
 *  as published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 * $Log: buff.c,v $
 * Revision 1.3  1999/07/24 16:55:28  francis
 * Adding support for entity-valued attributes
 *
 * Revision 1.2  1999/07/24 14:53:14  francis
 * Adding support for attribute value normalization
 *
 * Revision 1.1  1999/07/18 20:25:06  francis
 * Moving the global buff (static in parse.c) into the doxml_context,
 * to make the parser reentrant, so that it can be used from two threads
 * simultaneously.  Documenting the change; updating the version number
 * of the manual.
 *
 */

static const char* id="$Id: buff.c,v 1.3 1999/07/24 16:55:28 francis Exp $";

#include <doxml/buff.h>
#include <doxml/priv.h>

static const int buff_block_size=256;

int doxml_buff_init(doxml_context* context)
{
#ifdef DEBUG
  fprintf(stderr,"doxml_buff_init()\n");
#endif

  if (context->buff.data)
    free(context->buff.data);
  context->buff.data=(char*)malloc(buff_block_size);
  if (!context->buff.data)
    {
      fprintf(stderr,"doxml_buff_init(): malloc() failed\n");
      context->buff.size=context->buff.actual=0;
      return -1;
    }

  context->buff.size=buff_block_size;
  context->buff.actual=0;
  context->buff.last_ch=0;
  context->buff.deferred_space=0;

  return 0;
}

int doxml_buff_append(doxml_context* context,
                      char ch)
{
  return doxml_buff_append_collapse_space(context,ch,0);
}

int doxml_buff_append_str(doxml_context* context,
                          const char* s)
{
  return doxml_buff_append_str_collapse_space(context,s,0);
}

static int append_primitive(doxml_context* context,
                            char ch)
{

  if (context->buff.actual>context->buff.size)
    {
      fprintf(stderr,"doxml: append_primitive: buffer corrupted\n");
      return -1;
    }

  if (context->buff.actual==context->buff.size)
    {
      char* new_data=(char*)realloc(context->buff.data,context->buff.size+buff_block_size);
      if (!new_data)
        {
          fprintf(stderr,"doxml: append_primitive: can't grow buffer\n");
          return -1;
        }
      
      context->buff.data=new_data;
      context->buff.size+=buff_block_size;
    }
  
  context->buff.data[context->buff.actual++]=ch;
  return 0;
}

int doxml_buff_append_collapse_space(doxml_context* context,
                                     char ch,
                                     int collapse)
{
#ifdef DEBUG
  fprintf(stderr,"doxml_buff_append_collapse_space()\n");
#endif

  if (collapse)
    {
      if (ch==' ')
        {
          /* Leading space */
          if ((context->buff.actual)==0)
            return 0;

          /* Consecutive spaces */
          if ((context->buff.last_ch)==' ')
            return 0;

          context->buff.last_ch=' ';
          context->buff.deferred_space=1; /* Might be trailing */
          return 0;
        }
      else
        {
          if (context->buff.deferred_space)
            {
              if (append_primitive(context,' ')<0)
                {
                  fprintf(stderr,"doxml_buff_append_collapse_space(): can't append deferred space\n");
                  return -1;
                }
              context->buff.deferred_space=0;
            }
        }
    }

  context->buff.last_ch=ch;

  return append_primitive(context,ch);
}

int doxml_buff_append_str_collapse_space(doxml_context* context,
                                         const char* s,
                                         int collapse)
{
  if (!s)
    {
      fprintf(stderr,"doxml_buff_append_str(NULL)\n");
      return -1;
    }

  while (*s)
    {
      int err=doxml_buff_append_collapse_space(context,*(s++),collapse);
      if (err<0)
        return err;
    }

  return 0;
}

int doxml_buff_is_empty(doxml_context* context)
{
#ifdef DEBUG
  fprintf(stderr,"doxml_buff_is_empty()\n");
#endif

  return ((!context->buff.data) || (context->buff.actual==0));
}

const char* doxml_buff_extract(doxml_context* context)
{
  const char* res;

  if (doxml_buff_append(context,0)<0)
    {
      fprintf(stderr,"doxml_buff_extract(): can't append terminator to buffer\n");
      return 0;
    }

  res=context->buff.data;
  context->buff.data=0;
  context->buff.size=context->buff.actual=0;

#ifdef DEBUG
  fprintf(stderr,"doxml_buff_extract(): %s\n",res);
#endif

  return res;
}
