/*
 * This is part of the doxml library, a parser for XML.  It should not
 *  require any functionality not defined in ANSI C's standard libraries.
 *
 * Copyright (C) 1999 John Stracke, http://www.thibault.org
 *
 * This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public License
 *  as published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 * $Log: seekattr.c,v $
 * Revision 1.5  1999/07/31 15:14:32  francis
 * Fixing doxml_attr_seek2() to use compare_prefix_and_name
 * instead of compare_space_and_name
 *
 * Revision 1.4  1999/07/24 15:23:03  francis
 * Adding support for default attribute values
 *
 * Revision 1.3  1999/07/19 01:31:12  francis
 * Implementing namespace reform (but not using it yet).  Not documented yet.
 *
 * Revision 1.2  1999/06/07 04:57:38  francis
 * Implementing processing instructions.  Note the API changes.
 *
 * Revision 1.1  1999/06/01 01:51:14  francis
 * Adding functions to get/set/clear attributes.  Reorganzing the
 * to-do list.  Fixing a parser bug introduced I removed support for
 * attributes with no values at release 0.2.
 *
 */

static const char* id="$Id: seekattr.c,v 1.5 1999/07/31 15:14:32 francis Exp $";

#include <doxml/priv.h>

typedef int (*comparator)(const doxml_attribute* attr,
                          const void* context);

static doxml_attribute*
_doxml_attr_seek(doxml_atom* atom,
                 comparator comp,
                 const void* context)

{
  if (!atom)
    {
      fprintf(stderr,"_doxml_attr_seek(): atom==NULL\n");
      return 0;
    }

  if (!comp)
    {
      fprintf(stderr,"_doxml_attr_seek(): comp==NULL\n");
      return 0;
    }

  if ((atom->type)!=doxml_atom_element)
    {
      fprintf(stderr,"_doxml_attr_seek(): atom is not an element\n");
      return 0;
    }

  {
    doxml_attribute* cur;
    for (cur=atom->data.element.attrs.first; cur; cur=cur->next)
      {
        if ((*comp)(cur,context))
          return cur;
      }
  }

  return 0;
}

typedef struct {
  const char* namespace;
  const char* name;
} space_and_name;

static int
compare_space_and_name(const doxml_attribute* attr,
                       const void* context)
{
  const space_and_name* san=(const space_and_name*)context;
  if (!(attr && context))
    return 0;

  if (!(san->name))
    return 0;

  if (san->namespace)
    {
      if (!(attr->name.space))
        return 0;
      if (strcmp(san->namespace,attr->name.space))
        return 0;
    }
  else
    {
      if (attr->name.space)
        return 0;
    }
  
  if (!(attr->name.name)) /* can't happen */
    return 0;
  
  if (strcmp(san->name,attr->name.name))
    return 0;
  
  return 1;
}

static int
compare_prefix_and_name(const doxml_attribute* attr,
                        const void* context)
{
  const doxml_name_in_dtd* dtd_name=(const doxml_name_in_dtd*)context;
  if (!(attr && dtd_name))
    return 0;

  if (!(dtd_name->name))
    return 0;

  if (dtd_name->prefix)
    {
      if (!(attr->name.prefix))
        return 0;
      if (strcmp(dtd_name->prefix,attr->name.prefix))
        return 0;
    }
  else
    {
      if (attr->name.prefix)
        return 0;
    }
  
  if (!(attr->name.name)) /* can't happen */
    return 0;
  
  if (strcmp(dtd_name->name,attr->name.name))
    return 0;
  
  return 1;
}

doxml_attribute* doxml_attr_seek(doxml_atom* atom,
                                 const char* namespace,
                                 const char* name)
{
  space_and_name san;

  if (!name)
    {
      fprintf(stderr,"doxml_attr_seek(): name==NULL\n");
      return 0;
    }

  san.namespace=namespace;
  san.name=name;

  return _doxml_attr_seek(atom,compare_space_and_name,&san);
}

doxml_attribute* doxml_attr_seek2(doxml_atom* atom,
                                  const doxml_name_in_dtd* name)
{
  if (!name)
    {
      fprintf(stderr,"doxml_attr_seek(): name==NULL\n");
      return 0;
    }

  return _doxml_attr_seek(atom,compare_prefix_and_name,name);
}
