/*
 * This is part of the doxml library, a parser for XML.  It should not
 *  require any functionality not defined in ANSI C's standard libraries.
 *
 * This module implements doxml_register_uri_handler(),
 *  doxml_register_uri_handler_with_compound(), and
 *  doxml_unregister_uri_handler().  I know the GNU coding standards
 *  recommend not putting functions in the same module unless it's
 *  reasonable to expect that anybody who uses one will use another.
 *  doxml_register_uri_handler() calls
 *  doxml_register_uri_handler_with_compound(), so those two definitely
 *  go together; and I felt that it was safest to have
 *  doxml_unregister_uri_handler() here, too, so that both the functions
 *  that manipulate the linked list of handlers would be in the same
 *  place, and thus easier to maintain together.
 *
 * Copyright (C) 1999 John Stracke, http://www.thibault.org
 *
 * This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public License
 *  as published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 * $Log: urireg.c,v $
 * Revision 1.1  1999/06/25 21:15:01  francis
 * Adding a framework for the application to provide doxml with URI
 * handlers for use in processing external IDs.  (Not providing URI
 * code as part of doxml means we can work with whatever URI code the
 * app wants, and don't have to write anything non-portable.)
 *
 * Also, changing old _LIBXML_ stuff in the headers to _DOXML_.
 *
 */

static const char* id="$Id: urireg.c,v 1.1 1999/06/25 21:15:01 francis Exp $";

#include <doxml/uripriv.h>
#include <doxml/priv.h>

doxml_uri_handler_ref*
doxml_register_uri_handler(doxml_context* context,
                           doxml_uri_handler handler,
                           void* handler_context)
{
  if (!context)
    {
      fprintf(stderr,"doxml_register_uri_handler(): context is NULL\n");
      return 0;
    }

  if (!handler)
    {
      fprintf(stderr,"doxml_register_uri_handler(): handler is NULL\n");
      return 0;
    }

  if (!(context->compound_uri_handler))
    {
      context->compound_uri_handler=doxml_create_compound_uri_handler();
      if (!(context->compound_uri_handler))
        {
          fprintf(stderr,"doxml_register_uri_handler(): can't create compound for context\n");
          return 0;
        }
    }

  return
    doxml_register_uri_handler_with_compound(context->compound_uri_handler,
                                             handler,
                                             handler_context);
}

doxml_uri_handler_ref*
doxml_register_uri_handler_with_compound(doxml_compound_uri_handler* compound,
                                         doxml_uri_handler handler,
                                         void* handler_context)
{
  doxml_uri_handler_ref* res=0;

  if (!compound)
    {
      fprintf(stderr,
              "doxml_register_uri_handler_with_compound(): NULL compound\n");
      return 0;
    }

  if (!handler)
    {
      fprintf(stderr,
              "doxml_register_uri_handler_with_compound(): NULL handler\n");
      return 0;
    }

  res=(doxml_uri_handler_ref*)malloc(sizeof(doxml_uri_handler_ref));
  if (!res)
    {
      fprintf(stderr,
              "doxml_register_uri_handler_with_compound(): can't alloc res\n");
      return 0;
    }

  res->container.first=&(compound->first);
  res->container.last=&(compound->last);

  if (res->next=compound->first)
    res->next->prev=res;
  else
    compound->first=res;

  if (res->prev=compound->last)
    res->prev->next=res;
  else
    compound->last=res;

  res->handler=handler;
  res->handler_context=handler_context;

  return res;
}

void
doxml_unregister_uri_handler(doxml_uri_handler_ref* ref)
{
  if (!(ref && ref->container.first && ref->container.last))
    return;

  if (ref->next)
    ref->next->prev=ref->prev;
  else
    (*(ref->container.last))=ref->prev;

  if (ref->prev)
    ref->prev->next=ref->next;
  else
    (*(ref->container.first))=ref->next;

  free(ref);
}
