/*
 * This is part of the doxml library, a parser for XML.  It should not
 *  require any functionality not defined in ANSI C's standard libraries.
 *
 * Copyright (C) 1999 John Stracke, http://www.thibault.org
 *
 * This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public License
 *  as published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 * $Log: write.c,v $
 * Revision 1.27  1999/07/31 15:15:26  francis
 * Fixing it so that entity references in entity definitions get
 * written as themselves, instead of having the &s get replaced with
 * &amp;.
 *
 * Revision 1.26  1999/07/24 20:25:52  francis
 * Adding a comment
 *
 * Revision 1.25  1999/07/24 19:13:32  francis
 * Adding support for notations
 *
 * Revision 1.24  1999/07/24 16:55:29  francis
 * Adding support for entity-valued attributes
 *
 * Revision 1.23  1999/07/24 15:36:51  francis
 * Fixing a bug that was causing some namespaced attributes to be
 * printed without prefixes
 *
 * Revision 1.22  1999/07/24 15:28:57  francis
 * Fixing a bug that was causing some namespaced elements to be
 * printed without prefixes
 *
 * Revision 1.21  1999/07/19 01:31:12  francis
 * Implementing namespace reform (but not using it yet).  Not documented yet.
 *
 * Revision 1.20  1999/07/08 14:13:24  francis
 * Fixing a bug: when printing attributes, I wasn't including the
 * = (I have no idea how I missed this one--maybe it was introduced
 * when I switched to using write_in_quotes()).
 *
 * Revision 1.19  1999/06/25 19:43:03  francis
 * Adding support for unparsed entities
 *
 * Revision 1.18  1999/06/21 03:19:56  francis
 * Finishing up <!ENTITY> support
 *
 * Revision 1.17  1999/06/21 02:11:00  francis
 * Making sure default values in an <!ATTLIST> expand their & entities.
 * Also, in making those entities print out right, generalized the
 * escaping used in printing normal attributes
 *
 * Revision 1.16  1999/06/21 00:59:31  francis
 * Adding support for <!ATTLIST>s
 *
 * Revision 1.15  1999/06/20 22:23:54  francis
 * Adding support for <!ELEMENT>s in the doctype
 *
 * Revision 1.14  1999/06/20 20:05:34  francis
 * Adding support for PIs in the doctype
 *
 * Revision 1.13  1999/06/20 16:36:47  francis
 * Making doxml_write_document() use single quotes around the attributeoids
 * in the doctype's external-id if necessary.
 *
 * Revision 1.12  1999/06/20 16:30:14  francis
 * Adding partial support for <!DOCTYPE>.  It now handles doctypes with
 * names (all doctypes have names), with external-ids, but not with
 * any of the declarations.
 *
 * Revision 1.11  1999/06/08 03:00:19  francis
 * Writing entities where necessary (e.g., &quot; in attr values)
 *
 * Revision 1.10  1999/06/07 04:57:38  francis
 * Implementing processing instructions.  Note the API changes.
 *
 * Revision 1.9  1999/06/01 03:40:51  francis
 * Making sure that attrs which inherit their elements' namespaces don't
 * get written with prefixes.
 *
 * Revision 1.8  1999/06/01 03:37:44  francis
 * Making sure that elements which inherit their parents' namespaces don't
 * get written with prefixes.
 *
 * Revision 1.7  1999/06/01 03:29:42  francis
 * Minor style tweak
 *
 * Revision 1.6  1999/06/01 02:23:48  francis
 * Adding support for single quotes on attribute values
 *
 * Revision 1.5  1999/06/01 01:49:50  francis
 * Fixing a bug: the terminator tag was being written without a namespace.
 *
 * Revision 1.4  1999/05/21 23:10:29  francis
 * Reformatting, since renaming to doxml changed the widths of the
 * identifiers.
 *
 * Revision 1.3  1999/05/21 23:05:42  francis
 * Finished renaming to doxml
 *
 * Revision 1.2  1999/05/21 22:59:49  francis
 * Partially renaming to doxml
 *
 * Revision 1.1.1.1  1999/05/21 22:20:36  francis
 * Initial checkin of libxml, to be renamed doxml
 *
 */

static const char* id="$Id: write.c,v 1.27 1999/07/31 15:15:26 francis Exp $";

#include <string.h>

#include <doxml/priv.h>

static void
printEscaped(FILE* stream,
             const char* s,
             int escape_quot,
	     int escape_amp);

static void
write_in_quotes(FILE* stream,
                const char* str,
		int escape_amp)
{
  int escape_quot=0;
  char safe_quote='"';
  if (strchr(str,'"'))
    {
      if (strchr(str,'\''))
        escape_quot=1;
      else
        safe_quote='\'';
    }

  fprintf(stream,"%c",safe_quote);
  printEscaped(stream,str,escape_quot,escape_amp);
  fprintf(stream,"%c",safe_quote);
}


static void
write_spaces(FILE* stream,
             int num_spaces)
{
  static const char spaces[]="                                        ";
  int N=sizeof(spaces)-1;

  while (num_spaces>N)
    {
      fputs(spaces,stream);
      num_spaces-=N;
    }

  fputs(spaces+(N-num_spaces),stream);
}

static void
printEscaped(FILE* stream,
             const char* s,
             int escape_quot,
	     int escape_amp /* In an entity declaration, entity references
			     *  may appear, and must be printed as themselves.
			     */
	     )
{
  if (!(stream && s))
    return;

  while (*s)
    {
      const char* entity=0;
      switch (*s)
        {
        case '&':
	  if (escape_amp)
	    entity="amp";
          break;

/* Not needed because _doxml_write_atom() arranges the quotes so that it
 *  never needs to escape '.  Still in here just as a marker, to make
 *  it clear why it isn't done.
 */
#if 0
        case '\'':
          entity="apos";
          break;
#endif

        case '<':
          entity="lt";
          break;

        case '>':
          entity="gt";
          break;

        case '"':
          if (escape_quot)
            entity="quot";
          break;
        }

      if (entity)
        fprintf(stream,"&%s;",entity);
      else
        fputc(*s,stream);
      s++;
    }
}

static int
write_markupdecl_element_cp(FILE* stream,
                            doxml_markupdecl_element_cp* cp)
{
  char separator=0;
  doxml_markupdecl_element_cp* cur=0;
  int is_first=1;

  if (!stream)
    {
      fprintf(stderr,"doxml: write_markupdecl_element_cp(): stream is NULL\n");
      return -1;
    }

  if (!cp)
    {
      fprintf(stderr,"doxml: write_markupdecl_element_cp(): cp is NULL\n");
      return -1;
    }

  switch (cp->type)
    {
    case doxml_markupdecl_element_cp_name:
      if (cp->data.name.prefix)
        {
          fprintf(stream,"%s:",cp->data.name.prefix);
        }
      fprintf(stream,"%s",cp->data.name.name);
      goto multiplicity;

    case doxml_markupdecl_element_cp_choice:
      separator='|';
      cur=cp->data.choice;
      break;

    case doxml_markupdecl_element_cp_seq:
      separator=',';
      cur=cp->data.seq;
      break;

    case doxml_markupdecl_element_cp_invalid:
    default:
      fprintf(stderr,"doxml: write_markupdecl_element_cp(): invalid type\n");
      return -1;
    }

  if (!separator)
    return -1;

  fprintf(stream,"(");
  while (cur)
    {
      if (is_first)
        is_first=0;
      else
        fprintf(stream,"%c ",separator);

      if (write_markupdecl_element_cp(stream,cur)<0)
        {
          fprintf(stderr,"doxml: write_markupdecl_element_cp(): recursive call failed\n");
          return -1;
        }
      cur=cur->next;
    }
  fprintf(stream,")");

 multiplicity:
  {
    char multiplicity_char=0;
    switch (cp->multiplicity)
    {
    case doxml_multiplicity_1:
      break;

    case doxml_multiplicity_1_or_0:
      multiplicity_char='?';
      break;

    case doxml_multiplicity_0_or_more:
      multiplicity_char='*';
      break;

    case doxml_multiplicity_1_or_more:
      multiplicity_char='+';
      break;

    default:
      fprintf(stderr,
              "doxml: write_markupdecl_element_cp(): unknown multiplicity\n");
      return -1;
    }

    if (multiplicity_char)
      fprintf(stream,"%c",multiplicity_char);
  }

  return 0;
}

static int
write_doctype(FILE* stream,
              doxml_doctype* doctype)
{
  if (!stream)
    {
      fprintf(stderr,"doxml: write_doctype(): stream is NULL\n");
      return -1;
    }

  if (!doctype)
    {
      fprintf(stderr,"doxml: write_doctype(): doctype is NULL\n");
      return -1;
    }

  if (!(doctype->name.name))
    {
      fprintf(stderr,"doxml: write_doctype(): doctype->name.name is NULL\n");
      return -1;
    }

  fprintf(stream,"<!DOCTYPE ");
  if (doctype->name.prefix)
    fprintf(stream,"%s:",doctype->name.prefix);
  fprintf(stream,"%s",doctype->name);

  if (doctype->external_id)
    {
      fprintf(stream," %s",
              (doctype->external_id->is_public ? "PUBLIC" : "SYSTEM")
              );
      if (doctype->external_id->is_public)
        {
          if (!(doctype->external_id->pubid))
            {
              fprintf(stderr,"doxml: write_doctype(): external_id is_public, but pubid is NULL\n");
              return -1;
            }
          {
            char safe_quote='"';
            if (strchr(doctype->external_id->pubid,safe_quote))
              safe_quote='\'';
            fprintf(stream," %c%s%c",
                    safe_quote,
                    doctype->external_id->pubid,
                    safe_quote);
          }
        }

      if (!(doctype->external_id->system_uri))
        {
          fprintf(stderr,"doxml: write_doctype(): external_id->system_uri is NULL\n");
          return -1;
        }

      fprintf(stream," ");
      write_in_quotes(stream,doctype->external_id->system_uri,1);
    }

  if (doctype->decls.first)
    {
      doxml_markupdecl* cur;
      fprintf(stream,"\n [\n");
      for (cur=doctype->decls.first; cur; cur=cur->next)
        {
          switch (cur->type)
            {
            case doxml_markupdecl_element:
              fprintf(stream,"  <!ELEMENT ");

              if (cur->data.element.name.prefix)
                fprintf(stream,"%s:",cur->data.element.name.prefix);

              fprintf(stream,"%s ",cur->data.element.name.name);

              switch (cur->data.element.type)
                {
                case doxml_markupdecl_element_empty:
                  fprintf(stream,"EMPTY");
                  break;
                  
                case doxml_markupdecl_element_any:
                  fprintf(stream,"ANY");
                  break;
                  
                case doxml_markupdecl_element_mixed:
                  fprintf(stream,"(#PCDATA");
                  {
                    doxml_name_node* cur_name;
                    for (cur_name=cur->data.element.data.mixed;
                         cur_name;
                         cur_name=cur_name->next)
                      {
                        fprintf(stream," | ");
                        if (cur_name->name.prefix)
                          fprintf(stream,"%s:",cur_name->name.prefix);
                        if (!(cur_name->name.name))
                          {
                            fprintf(stderr,"doxml: write_doctype(): NULL name in mixed <!ELEMENT>\n");
                            return -1;
                          }
                        fprintf(stream,"%s",cur_name->name.name);
                      }

                    fprintf(stream,")");
                    if (cur->data.element.data.mixed)
                      fprintf(stream,"*");
                  }
                  break;
                  
                case doxml_markupdecl_element_children:
                  if (write_markupdecl_element_cp(stream,
                                                  cur->data.element.data.children)
                      <0)
                    {
                      fprintf(stderr,"doxml: write_doctype(): write_markupdecl_element_cp() failed\n");
                      return -1;
                    }
                  break;

                default:
                  fprintf(stderr,"doxml: write_doctype(): unknown element type\n");
                  break;
                }
              fprintf(stream,">\n");
              break;
              
            case doxml_markupdecl_attlist:
              fprintf(stream,"  <!ATTLIST ");
              if (cur->data.attlist.name.prefix)
                fprintf(stream,"%s:",cur->data.attlist.name.prefix);

              if (!(cur->data.attlist.name.name))
                {
                  fprintf(stderr,
                          "doxml: write_doctype(): ATTLIST with no name\n");
                  return -1;
                }

              fprintf(stream,"%s",cur->data.attlist.name.name);
              if (cur->data.attlist.attdefs)
                {
                  doxml_markupdecl_attdef* cur_def;
                  for (cur_def=cur->data.attlist.attdefs;
                       cur_def;
                       cur_def=cur_def->next
                       )
                    {
                      const char* type_str=0;

                      fprintf(stream,"\n            ");
                      if (cur_def->name.prefix)
                        fprintf(stream,"%s:",cur_def->name.prefix);
                      
                      if (!(cur_def->name.name))
                        {
                          fprintf(stderr,
                                  "doxml: write_doctype(): attdef with no name\n");
                          return -1;
                        }

                      fprintf(stream,"%s ",cur_def->name.name);
                      switch (cur_def->attype.type)
                        {
                        case doxml_markupdecl_attdef_attype_cdata:
                          fprintf(stream,"CDATA ");
                          break;
                          
                        case doxml_markupdecl_attdef_attype_id:
                          fprintf(stream,"ID ");
                          break;
                          
                        case doxml_markupdecl_attdef_attype_idref:
                          fprintf(stream,"IDREF ");
                          break;
                          
                        case doxml_markupdecl_attdef_attype_idrefs:
                          fprintf(stream,"IDREFS ");
                          break;
                          
                        case doxml_markupdecl_attdef_attype_entity:
                          fprintf(stream,"ENTITY ");
                          break;
                          
                        case doxml_markupdecl_attdef_attype_entities:
                          fprintf(stream,"ENTITIES ");
                          break;
                          
                        case doxml_markupdecl_attdef_attype_nmtoken:
                          fprintf(stream,"NMTOKEN ");
                          break;
                          
                        case doxml_markupdecl_attdef_attype_nmtokens:
                          fprintf(stream,"NMTOKENS ");
                          break;
                          
                        case doxml_markupdecl_attdef_attype_enumerated:
                          if (cur_def->attype.enumeration.is_notation)
                            fprintf(stream,"NOTATION ");
                          
                          fprintf(stream,"(");
                          {
                            int is_first=1;
                            doxml_name_node* cur_name;
                            for (cur_name=cur_def->attype.enumeration.names;
                                 cur_name;
                                 cur_name=cur_name->next
                                 )
                              {
                                if (is_first)
                                  is_first=0;
                                else
                                  fprintf(stream," | ");

                                if (cur_name->name.prefix)
                                  fprintf(stream,"%s:",cur_name->name.prefix);

                                if (!(cur_name->name.name))
                                  {
                                    fprintf(stderr,
                                            "doxml: write_doctype(): attdef enumeration item with no name\n");
                                    return -1;
                                  }

                                fprintf(stream,cur_name->name.name);
                              }
                          }
                          fprintf(stream,") ");
                          break;
                        }

                      switch (cur_def->def.type)
                        {
                        case doxml_markupdecl_attdef_default_required:
                          fprintf(stream,"#REQUIRED");
                          break;
                          
                        case doxml_markupdecl_attdef_default_implied:
                          fprintf(stream,"#IMPLIED");
                          break;
                          
                        case doxml_markupdecl_attdef_default_fixed:
                          fprintf(stream,"#FIXED ");
                          
                        case doxml_markupdecl_attdef_default_plain:
                          if (!(cur_def->def.value))
                            {
                              fprintf(stderr,
                                      "doxml: write_doctype(): attdef default value NULL\n");
                              return -1;
                            }
                          write_in_quotes(stream,cur_def->def.value,1);
                          break;

                        default:
                          fprintf(stderr,"doxml: write_doctype(): unknown default type\n");
                          return -1;
                        }
                    }
                  fprintf(stream,"\n   >\n");
                }
              else
                fprintf(stream,">\n");
              break;
              
            case doxml_markupdecl_entity:
              if (!(cur->data.entity.name.name))
                {
                  fprintf(stderr,
                          "doxml: write_doctype(): ENTITY with no name\n");
                  return -1;
                }
              fprintf(stream,"  <!ENTITY ");
              if (!(cur->data.entity.is_general))
                fprintf(stream,"%% ");
              if (cur->data.entity.name.prefix)
                fprintf(stream,"%s:",cur->data.entity.name.prefix);
              fprintf(stream,"%s ",cur->data.entity.name.name);
              if (cur->data.entity.is_general)
                {
                  if (cur->data.entity.data.general.is_external)
                    {
                      fprintf(stream,"%s ",
                              (cur->data.entity.data.general.data.external.external_id.is_public ? "PUBLIC" : "SYSTEM")
                              );
                      if (cur->data.entity.data.general.data.external.external_id.is_public)
                        {
                          if (!(cur->data.entity.data.general.data.external.external_id.pubid))
                            {
                              fprintf(stderr,"doxml: write_doctype(): external_id is_public, but pubid is NULL\n");
                              return -1;
                            }
                          {
                            write_in_quotes(stream,cur->data.entity.data.general.data.external.external_id.pubid,
					    1);
                            fprintf(stream," ");
                          }
                        }
                      
                      if (!(cur->data.entity.data.general.data.external.external_id.system_uri))
                        {
                          fprintf(stderr,"doxml: write_doctype(): external_id->system_uri is NULL\n");
                          return -1;
                        }
                      
                      write_in_quotes(stream,
                                      cur->data.entity.data.general.data.external.external_id.system_uri,
				      1);

                      if (cur->data.entity.data.general.data.external.notation_name.name)
                        {
                          fprintf(stream," NDATA ");
                          if (cur->data.entity.data.general.data.external.notation_name.prefix)
                            fprintf(stderr,"%s:",cur->data.entity.data.general.data.external.notation_name.prefix);
                          fprintf(stream,"%s",
                                  cur->data.entity.data.general.data.external.notation_name.name
                                  );
                        }
                    }
                  else /* not external */
                    write_in_quotes(stream,
                                    cur->data.entity.data.general.data.internal,
				    0);
                }
              else /* not general */
                {
                  if (cur->data.entity.data.param.is_external)
                    {
                      fprintf(stream,"%s ",
                              (cur->data.entity.data.param.data.external_id.is_public ? "PUBLIC" : "SYSTEM")
                              );
                      if (cur->data.entity.data.param.data.external_id.is_public)
                        {
                          if (!(cur->data.entity.data.param.data.external_id.pubid))
                            {
                              fprintf(stderr,"doxml: write_doctype(): external_id is_public, but pubid is NULL\n");
                              return -1;
                            }
                          {
                            write_in_quotes(stream,cur->data.entity.data.param.data.external_id.pubid,
					    1);
                            fprintf(stream," ");
                          }
                        }
                      
                      if (!(cur->data.entity.data.param.data.external_id.system_uri))
                        {
                          fprintf(stderr,"doxml: write_doctype(): external_id->system_uri is NULL\n");
                          return -1;
                        }
                      
                      write_in_quotes(stream,
                                      cur->data.entity.data.param.data.external_id.system_uri,
				      1);
                    }
                  else /* not external */
                    write_in_quotes(stream,
                                    cur->data.entity.data.param.data.internal,
				    0);
                }
              fprintf(stream,">\n");
              break;
              
            case doxml_markupdecl_notation:
              fprintf(stream,"  <!NOTATION ");

              if (cur->data.notation.name.prefix)
                fprintf(stream,"%s:",cur->data.notation.name.prefix);

              if (cur->data.notation.name.name)
                fprintf(stream,"%s",cur->data.notation.name.name);
              else
                {
                  fprintf(stderr,"doxml: write_doctype(): NOTATION with no name\n");
                  return -1;
                }

              {
                doxml_external_id* external_id
                  =&(cur->data.notation.external_id);
                if (external_id->is_public)
                  {
                    if (external_id->pubid)
                      {
                        fprintf(stream," PUBLIC");
                        write_in_quotes(stream,external_id->pubid,1);
                      }
                    else
                      {
                        fprintf(stderr,
                                "doxml: write_doctype(): PUBLIC NOTATION with no pubid\n");
                        return -1;
                      }
                  }
                else
                  fprintf(stream," SYSTEM ");

                if (external_id->system_uri)
                  write_in_quotes(stream,external_id->system_uri,1);
                else
                  {
                    fprintf(stderr,"doxml: write_doctype(): NOTATION with no system_uri\n");
                    return -1;
                  }

                fprintf(stream,">\n");
              }
              break;
              
            case doxml_markupdecl_pi:
              if (!cur->data.pi.name.name)
                {
                  fprintf(stderr,"doxml: write_doctype(): PI with no name\n");
                  return -1;
                }
              fprintf(stream,"  <?");
              if (cur->data.pi.name.prefix)
                fprintf(stream,"%s:",cur->data.pi.name.prefix);
              if (cur->data.pi.value)
                fprintf(stream,"%s?>\n",
                        cur->data.pi.name,
                        cur->data.pi.value);
              else
                fprintf(stream,"%s?>\n",
                        cur->data.pi.name);
              break;
              
            case doxml_markupdecl_pe_reference:
              /* This is not something that needs to be supported, since
               *  a PEreference will get expanded by the parser, so
               *  we'll never see one in the parsed data.
               */
              fprintf(stream,"  [pe_reference unsupported]\n");
              break;

            default:
              fprintf(stderr,"doxml: write_doctype(): unknown markup type\n");
              return -1;
            }
        }
      fprintf(stream," ]");
    }

  fprintf(stream,">\n");

  return 0;
}

static int
_doxml_write_atom(FILE* stream,
                  doxml_atom* atom,
                  int indent)
{
  write_spaces(stream,indent);
  switch (atom->type)
    {
    case doxml_atom_text:
      {
        const char* s=atom->data.text.text;

        printEscaped(stream,atom->data.text.text,0,1);
        fputc('\n',stream);
      }
      break;

    case doxml_atom_element:
      {
        fputs("<",stream);
        {
          const char* prefix=atom->data.element.name.prefix;
          if (!prefix)
            {
              if (atom->data.element.name.space)
                {
                  if (!(   (atom->parent)
                        && (atom->parent->data.element.name.space)
                        && (!strcmp(atom->parent->data.element.name.space,
                                    atom->data.element.name.space
                                    )
                            )
                        )
                      )
                    {
                      prefix=doxml_lookup_prefix(atom,
                                                 atom->data.element.name.space);
                    }
                }
            }
          if (prefix)
            fprintf(stream,"%s:",prefix);
        }
        
        if (atom->data.element.name.name)
          fputs(atom->data.element.name.name,stream);
        else
          {
            fprintf(stderr,"doxml: _doxml_write_atom(): element's tag is NULL\n");
            return -1;
          }

        if (atom->data.element.attrs.first)
          {
            doxml_attribute* cur=atom->data.element.attrs.first;

            fputc('\n',stream);
          
            do
              {
                write_spaces(stream,indent+1);
                {
                  const char* prefix=cur->name.prefix;
                  if (!prefix)
                    {
                      if (cur->name.space)
                        {
                          if (!(   atom->data.element.name.space
                                && (!(strcmp(atom->data.element.name.space,
                                             cur->name.space
                                             )
                                      )
                                    )
                                )
                              )
                            {
                              prefix=doxml_lookup_prefix(atom,
                                                         cur->name.space);
                            }
                        }
                    }
                  if (prefix)
                    fprintf(stream,"%s:",prefix);
                }

                if (cur->name.name)
                  fputs(cur->name.name,stream);
                else
                  {
                    fprintf(stderr,"doxml: _doxml_write_atom(): attr's name is NULL\n");
                    return -1;
                  }

                if (cur->is_entity)
                  {
                    if (cur->value.entity)
                      {
                        const doxml_name_in_dtd* name
                          =&(cur->value.entity->data.entity.name);
                        fputc('=',stream);
                        if (name->prefix)
                          {
                            char* fullname=(char*)malloc(strlen(name->prefix)
                                                         +1
                                                         +strlen(name->name)
                                                         +1
                                                         );
                            if (fullname)
                              {
                                sprintf(fullname,"%s:%s",
                                        name->prefix,name->name);
                                write_in_quotes(stream,fullname,1);
                                free(fullname);
                              }
                          }
                        else
                          write_in_quotes(stream,name->name,1);
                      }
                  }
                else
                  {
                    if (cur->value.literal)
                      {
                        fputc('=',stream);
                        write_in_quotes(stream,cur->value.literal,1);
                      }
                  }

                fputc('\n',stream);
              }
            while (cur=cur->next);

            write_spaces(stream,indent+1);
          }

        if (!(atom->data.element.atoms.first))
          fputc('/',stream);

        fputs(">\n",stream);

        {
          doxml_atom* cur=atom->data.element.atoms.first;
          while (cur)
            {
              if (_doxml_write_atom(stream,cur,indent+1)<0)
                return -1;
              cur=cur->next;
            }
        }

        if (atom->data.element.atoms.first)
          {
            write_spaces(stream,indent);
            fprintf(stream,"</");
            if (atom->data.element.name.space)
              {
                const char* prefix=atom->data.element.name.prefix;
                if (!prefix)
                  prefix=doxml_lookup_prefix(atom,
                                             atom->data.element.name.space);
                if (prefix)
                  fprintf(stream,"%s:",prefix);
              }
            fprintf(stream,"%s>\n",atom->data.element.name.name);
          }
      }
      break;

    case doxml_atom_pi:
      if (atom->data.pi.name.name)
        {
          fprintf(stream,"<?");
          if (atom->data.pi.name.prefix)
            fprintf(stream,"%s:",atom->data.pi.name.prefix);
          fprintf(stream,"%s",atom->data.pi.name.name);
          if (atom->data.pi.value)
            fprintf(stream," %s",atom->data.pi.value);
          fprintf(stream,"?>\n");
        }
      break;

    case doxml_atom_external_entity:
      if (!(atom->data.external_entity.decl))
        {
          fprintf(stderr,"doxml: _doxml_write_atom(): external_entity type with NULL decl\n");
          return -1;
        }

      if ((atom->data.external_entity.decl->type)!=doxml_markupdecl_entity)
        {
          fprintf(stderr,"doxml: _doxml_write_atom(): external_entity type with non-entity decl\n");
          return -1;
        }

      if (!(atom->data.external_entity.decl->data.entity.name.name))
        {
          fprintf(stderr,"doxml: _doxml_write_atom(): external_entity with NULL decl name\n");
          return -1;
        }

      fprintf(stream," &");
      if (atom->data.external_entity.decl->data.entity.name.prefix)
        fprintf(stream,"%s:",
                atom->data.external_entity.decl->data.entity.name.prefix);
      fprintf(stream,"%s;\n",
              atom->data.external_entity.decl->data.entity.name.name);
      return 0;
    }
  
  return 0;
}

int doxml_write_atom(FILE* stream,
                     doxml_atom* atom)
{
  if (!atom)
    {
      fprintf(stderr,"doxml: doxml_write_atom(): atom is NULL\n");
      return -1;
    }

  if (!stream)
    {
      fprintf(stderr,"doxml: doxml_write_atom(): stream is NULL\n");
      return -1;
    }

  return _doxml_write_atom(stream,atom,0);
}

int doxml_write_document(FILE* stream,
                         doxml_document* document)
{
  if (!document)
    {
      fprintf(stderr,"doxml: doxml_write_document(): document is NULL\n");
      return -1;
    }

  if (!stream)
    {
      fprintf(stderr,"doxml: doxml_write_document(): stream is NULL\n");
      return -1;
    }

  fprintf(stream,"<?xml version=\"1.0\" encoding='UTF-8'?>\n");

  if (document->doctype)
    {
      if (write_doctype(stream,document->doctype)<0)
        {
          fprintf(stderr,
                  "doxml: doxml_write_document(): write_doctype() failed\n");
          return -1;
        }
    }

  if (document->prolog)
    {
      doxml_atom* cur=document->prolog;
      while (cur)
        {
          int err=doxml_write_atom(stream,cur);
          if (err<0)
            {
              fprintf(stderr,"doxml: doxml_write_document(): doxml_write_atom() failed\n");
              return err;
            }

          cur=cur->next;
        }
    }

  if (document->root)
    return doxml_write_atom(stream,document->root);
  else
    return 0;
}
